=== modified file 'bzrlib/branch.py'
--- bzrlib/branch.py	2005-11-13 02:11:30 +0000
+++ bzrlib/branch.py	2005-11-13 02:32:17 +0000
@@ -632,8 +632,12 @@
             # because we can't connect to the remote location.
             # How do we distinguish this from a remote branch
             # which has been deleted?
-            rev_history = self._update_remote_location(bound_loc,
-                    rev_history[-1])
+            try:
+                rev_history = self._update_remote_location(bound_loc,
+                                                           rev_history)
+            except DivergedBranches:
+                raise errors.CannotInstallRevisions('Remote tree has commits.'
+                            ' Use bzr pull to come up to date')
         self.put_controlfile('revision-history', '\n'.join(rev_history))
 
     def has_revision(self, revision_id):
@@ -789,7 +793,7 @@
         else:
             return None
 
-    def missing_revisions(self, other, stop_revision=None, diverged_ok=False):
+    def missing_revisions(self, other, stop_revision=None, other_history=None):
         """Return a list of new revisions that would perfectly fit.
         
         If self and other have not diverged, return a list of the revisions
@@ -819,7 +823,8 @@
         """
         self_history = self.revision_history()
         self_len = len(self_history)
-        other_history = other.revision_history()
+        if other_history is None:
+            other_history = other.revision_history()
         other_len = len(other_history)
         common_index = min(self_len, other_len) -1
         if common_index >= 0 and \
@@ -834,26 +839,43 @@
                 raise bzrlib.errors.NoSuchRevision(self, stop_revision)
         return other_history[self_len:stop_revision]
 
-    def update_revisions(self, other, stop_revision=None):
-        """Pull in new perfect-fit revisions."""
+    def update_revisions(self, other, stop_revision=None, other_history=None):
+        """Pull in new perfect-fit revisions.
+
+        :param other: Another Branch to pull from
+        :param stop_revision: Updated until the given revision
+        :param other_history: Alternative history to other.revision_history()
+        :return: None
+        """
         from bzrlib.fetch import greedy_fetch
         if stop_revision is None:
-            stop_revision = other.last_revision()
+            if other_history is not None:
+                stop_revision = other_history[-1]
+            else:
+                stop_revision = other.last_revision()
         ### Should this be checking is_ancestor instead of revision_history?
         if (stop_revision is not None and 
             stop_revision in self.revision_history()):
             return
         greedy_fetch(to_branch=self, from_branch=other,
                      revision=stop_revision)
-        pullable_revs = self.pullable_revisions(other, stop_revision)
+        pullable_revs = self.pullable_revisions(other, stop_revision,
+                other_history=other_history)
         if len(pullable_revs) > 0:
             self.append_revision(*pullable_revs)
 
-    def pullable_revisions(self, other, stop_revision):
-        try:
+    def pullable_revisions(self, other, stop_revision, other_history=None):
+        if other_history is not None:
+            try:
+                other_revno = other_history.index(stop_revision) + 1
+            except ValueError:
+                raise errors.NoSuchRevision(self, stop_revision)
+        else:
             other_revno = other.revision_id_to_revno(stop_revision)
-            return self.missing_revisions(other, other_revno)
-        except (DivergedBranches, errors.NoSuchRevision), e:
+        try:
+            return self.missing_revisions(other, other_revno,
+                    other_history=other_history)
+        except DivergedBranches, e:
             try:
                 pullable_revs = get_intervening_revisions(self.last_revision(),
                                                           stop_revision, self)
@@ -1166,20 +1188,15 @@
             pass
 
     @needs_read_lock
-    def _update_remote_location(self, other_loc, revision):
+    def _update_remote_location(self, other_loc, revision_history):
         """Make sure the remote location has the local changes.
 
         :param other_loc: Path to the other location
-        :param revision: Revision which needs to inserted
-                         into the remote tree. (plus ancestry)
+        :param revision_history: Total history to be updated
         :return: The remote revision_history
-
-        TODO: update_revisions is too expensive when we know 
-              what the real history should be most of the time,
-              as it has to trace ancestry.
         """
         from bzrlib.fetch import greedy_fetch
-        mutter('_update_remote_location: %r, %r', other_loc, revision)
+        mutter('_update_remote_location: %r, %r', other_loc, revision_history)
         other = Branch.open(other_loc)
         bound_loc = other.get_bound_location()
         if bound_loc is not None:
@@ -1187,7 +1204,7 @@
         other.lock_write()
         try:
             # update_revisions should also append to the revision history.
-            other.update_revisions(self, stop_revision=revision)
+            other.update_revisions(self, other_history=revision_history)
             return other.revision_history()
         finally:
             other.unlock()

=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2005-11-13 01:50:16 +0000
+++ bzrlib/builtins.py	2005-11-13 02:32:17 +0000
@@ -28,6 +28,7 @@
 from bzrlib.branch import Branch
 from bzrlib.errors import BzrError, BzrCheckError, BzrCommandError, NotBranchError
 from bzrlib.errors import DivergedBranches, NoSuchFile, NoWorkingTree
+import bzrlib.errors as errors
 from bzrlib.option import Option
 from bzrlib.revisionspec import RevisionSpec
 import bzrlib.trace
@@ -1231,6 +1232,8 @@
         except StrictCommitFailed:
             raise BzrCommandError("Commit refused because there are unknown "
                                   "files in the working tree.")
+        except errors.CannotInstallRevisions, e:
+            raise BzrCommandError(e.msg)
 
 
 class cmd_check(Command):

